module file

import Base.show

export FITFile, read_file, DataRecord

include("bytes.jl"); using .bytes
include("mesg_num.jl"); using .mesg_num
include("base_types.jl"); using .base_types

HEADER_SZ = 14

struct FITHeader
    sz::UInt8
    protocol::UInt8
    version::Integer
    size::Integer
    type::String
    crc::AbstractVector{<:Real}
end

@enum RecordType begin
    definition = 1
    data = 2
    timestamp = 3
end

"""
    RecordHeader

Contains the header byte for a record row
"""
mutable struct RecordHeader
    type::RecordType
    local_mesg_type::Int64
    time_offset::Union{Nothing, Int64}
end

struct FieldDefinition
    type::AbstractString
    sz_bytes::UInt8
    base_type::Type # TODO: This needs to be translated into an actual base type.
end

mutable struct DefinitionBody
    architecture::AbstractString
    global_mesg_type::AbstractString # Read from mesg_num.jl using the bytes in this field.
    num_fields::UInt8
    field_definitions::Vector{FieldDefinition}
end

"""
    LocalMesgTypeMapping

For use in mapping local message type to a struct containing useful info.
"""
struct LocalMesgTypeMapping
    row_sz::Int64
    global_mesg_type::AbstractString
end

mutable struct DataBody
end

"""
    DataRecord

Represents a DataRecord within the Garmin file.
"""
mutable struct DataRecord
    hdr::Union{RecordHeader, Nothing}
    body::Union{DefinitionBody, DataBody}
end

mutable struct FITFileReader
    filepath::AbstractString
    fp::Union{IO, Nothing}
    bytes_read::Int64
    header::Union{FITHeader, Nothing}
    body::Union{Vector{DataRecord}, Nothing}
    mesg_map::Dict{Real, LocalMesgTypeMapping}
    next_row_sz::Int64
end


blank_file(fp::AbstractString) = FITFileReader(fp, nothing, 0, nothing, nothing, Dict(), 0)
blank_row() = DataRecord(nothing)

"""
    open_fit_file!(f::FITFileReader)

Open the fit file, and set f.fp
"""
function open_fit_file!(f::FITFileReader)
    f.fp = open(f.filepath, "r")
end

"""
    close_fit_file!(f::FITFileReader)

Close the fit file, and set our fp to nothing.
"""
function close_fit_file!(f::FITFileReader)
    close(f.fp)
    f.fp = nothing
end

"""
    read_bytes!(f::FITFileReader, nb::Integer)::Vector{UInt8}

Read nb bytes from f, update f.bytes_read, return bytes
"""
function read_bytes!(f::FITFileReader, nb::Integer)::Vector{UInt8}
    bytes = read(f.fp, nb; all=false)
    f.bytes_read += max(length(bytes), nb)
    bytes
end

"""
    read_header(fit_file::IOStream)

Read the header of the given fit file, and return a FITHeader struct
"""
function read_header!(f::FITFileReader)
    header = read_bytes!(f, HEADER_SZ)
    f.header = FITHeader(
        header[1],
        header[2],
        byte_vec_to_int(header[3:4]),
        byte_vec_to_int(header[5:8]),
        byte_vec_to_string(header[9:12]),
        header[13:14]
    )
end

"""
    decode_header(v::Vector{UInt8})::RecordHeader

Decode the header byte, determine if the record is a definition, data, or timestamp.
"""
function decode_header(v::Vector{UInt8})::RecordHeader
    type = if v[8] == 1
        timestamp 
    elseif v[7] == 1
        definition
    else
        data
    end

    if type == timestamp
        local_mesg_type = sum_bits(v[6:7])
        time_offset = sum_bits(v[1:5])
        RecordHeader(type, local_mesg_type, time_offset)
    else
        local_mesg_type = sum_bits(v[1:4])
        RecordHeader(type, local_mesg_type, nothing)
    end
end

"""
    parse_field(bytes::Vector{UInt8})::FieldDefinition

Parse a field definition in the definition message into a struct
"""
function parse_field(mesg_type::AbstractString, bytes::Vector{UInt8})::FieldDefinition
    # TODO the first byte of the field definition, the field num, is definiedre-defined by each mesg definitionn
    FieldDefinition(
        get_field_description(mesg_type, bytes[1]),
        bytes[2],
        get_base_type(bytes[3])
    )
end

function get_row_size(field_definitions::Vector{FieldDefinition})::Int64
    sum([def.sz_bytes for def ∈ field_definitions])
end

"""
    read_record_content!(f::FITFileReader, hdr::RecordHeader)::Union{DefinitionBody, DataBody}

Using f, read the appropriate message body given the hdr (which tells us if this is def or data)
"""
function read_record_content!(f::FITFileReader, hdr::RecordHeader)::Union{DefinitionBody, DataBody}
    # Read the definition record, return the body
    if hdr.type == definition
        # Read all bytes first
        header_bytes = read_bytes!(f, 5)
        header_field_definitions = read_bytes!(f, header_bytes[end] * 3) # sz * field size 
        header_developer_flag = read_bytes!(f, 1)
        if header_developer_flag[1] > 0x00
            header_developer_fields = read_bytes!(f, header_developer_flag[1] * 3)
        else
            header_developer_fields::Vector{UInt8} = []
        end

        # Parse the bytes
        global_mesg_type = get_mesg_num_string(byte_vec_to_int(header_bytes[3:4]))
        field_definitions = [parse_field(global_mesg_type, header_field_definitions[i:i+2]) for i ∈ 1:3:length(header_field_definitions)]
        developer_field_definitions = [parse_field(global_mesg_type, header_developer_fields[1:1+2]) for i ∈ 1:3:length(header_developer_fields)]

        # Get the total row size from parsing the fields.
        row_sz = sum(get_row_size.([field_definitions, developer_field_definitions]))

        # Associate the global message type with this local message type.
        f.mesg_map[hdr.local_mesg_type] = LocalMesgTypeMapping(row_sz, global_mesg_type)

        DefinitionBody(
            header_bytes[2] == 0 ? "littleendian" : "bigendian",
            global_mesg_type,
            header_bytes[end],
            field_definitions
        )
    else
        row_bytes = read_bytes!(f, f.next_row_sz)
    end
end

"""
    read_records!(f::FITFileReader)

Read all the data record messages.
"""
function read_records!(f::FITFileReader)
    if f.bytes_read < 12 || f.bytes_read > 14
        throw("invalid header read, not within header size")
    elseif isnothing(f.header) 
        throw("cant read FIT file, no header")
    end

    # Loop through bytes_read until done
    while f.bytes_read < f.header.size
        # Read header byte.
        header_bits = bit_vector(read_bytes!(f, 1); little_endian=true)
        header = decode_header(header_bits)
        record = DataRecord(header, read_record_content!(f, header))

        if isnothing(f.body)
            f.body = [record]
        else
            push!(f.body, record)
        end
    end
end

function read_file(filepath::AbstractString)::FITFileReader
    f = blank_file(filepath)

    open_fit_file!(f)
    try
        read_header!(f)
        read_records!(f)
        # read_rows!(f)
    catch e
        rethrow(e)
    finally
        close_fit_file!(f)
    end
    f
end

function show(io::IO, hdr::RecordHeader)
    type = hdr.definition ? "Definition" : "Data"
    println(io, "Header($(type), local_message_type=$(hdr.local_mesg_type))")
end

function show(io::IO, ffr::FITFileReader)
    println(io, "FITFileReader($(ffr.filepath), bytes_read=$(ffr.bytes_read), header=$(ffr.header), body=$(ffr.body))")
end

function show(io::IO, defbody::DefinitionBody)
    println(io, "DefinitionBody(arch=$(defbody.architecture),")
    println(io, "               global_type=$(defbody.global_mesg_type)")
    println(io, "               num_fields=$(defbody.num_fields)")
    println(io, "               field_defs=$(defbody.field_definitions)")
end

end